import Simple_Process_REPL.core as r
from Simple_Process_REPL.options import define_help_text
import Particle_Board_REPL.particle as P
import regex as re
import logging
import os
import yaml


"""
The purpose of this module is to Serve as a liaison between the
Interpreter/REPL and the device interface layer, which in this case is
defined in particle.py

The Repl module requires two symbol tables, and a dictionary of
defaults, as well as any stateful data the application would like to keep.
In this case it is the _device_ dictionary below.

In many cases particle functions may be used directly in the symbol table.
In other cases where additional functionality,
or data use is needed, the function would be defined here.

The symbol table of these functions is defined and passed to
to the repl core module which is reponsible for similar things
as well as parsing the command line, setting up logging, and
executing the process as desired.
"""

# This is a map which is merged with the application state map.
# Defaults are used by the cli where appropriate.
# device is our particle board, we need it's id, and the
# the path of it's device, ie. /dev/ttyUSB..., The board
# name is boron, photon, whatever. and the last_id just
# in case we need it.

# Particle board interface state
PB = {
    "device": {"id": "", "name": "", "path": "", "last_id": ""},
    "defaults": {
        "config_file": "PBRconfig.yaml",
        "loglevel": "info",
        "logfile": "PBR.log",
    },
}

PB |= r.load_defaults("Particle_Board_REPL", "PBR-defaults.yaml")

logger = logging.getLogger()

prefix = """PBR, The Particle_Board_REPL, is an application
Built upon the Simple_Process_REPL. It is intended to make it easy
to interact, program and test Particle.io Boards.
Using the REPL with `PBR -r` is the easiest way to get to know
the features of this application. By default configurations files are
loaded so that basic functionalities are provided for. If a configuration
file is present it will be merged into the defaults. A configuration file
can be created with the `save_config` command.
"""

suffix = ""

define_help_text(prefix, suffix)


def see_images():
    """Ask if the test images are seen, fail if not."""
    yno_msg = r.get_in_config(["dialogs", "images_seen"])
    if not r.ynbox(yno_msg):
        raise Exception("Failure: Test images not Seen.")


def input_serial():
    """
    Function called by test handshake, Asks if images are seen,
    then presents a dialog loop to receive an 8 digit serial number
    to be returned back to the handshake and sent to the device.
    """
    see_images()

    msg = r.get_in_config(["dialogs", "input_serial"])
    while True:
        code, res = r.inputbox(
            msg, title=r.get_in_config(["dialogs", "title"]), height=10, width=50
        )
        if re.match(r"^\d{8}$", res):
            yno_msg = "%s : %s" % (
                r.get_in_config(["dialogs", "serial_is_correct"]),
                res,
            )
            if r.ynbox(yno_msg):
                break
        else:
            r.msgbox(r.get_in_config(["dialogs", "serial_must"]))

    os.system("clear")
    logger.info("Serial Number Entered is: %s" % res)
    return res


def flash_image():
    "Flash the flash image"
    P.flash(r.get_in_config(["images", "flash"]))


def flash_tinker():
    "Flash the tinker image"
    P.flash(r.get_in_config(["images", "tinker"]))


def flash_test():
    "Flash the test image"
    P.flash(r.get_in_config(["images", "test"]))


def product_add():
    """ product device add product id - associate the device wit a product"""
    P.product_add(r.get_in_config(["particle", "product"]), r.get_in_device("id"))


def claim():
    "Claim the device, cloud claim"
    P.claim(r.get_in_device("id"))


def add():
    "Claim the device, device add"
    P.add(r.get_in_device("id"))


def release():
    "Release the claim on a device."
    P.release(r.get_in_device("id"))


def cloud_status():
    "Check to see if the device is connected to the cloud"
    P.cloud_status(r.get_in_device("id"))


def reset_usb():
    "Reset the usb device."
    P.reset_usb(r.get_in_device("id"))


def list_usb_w_timeout():
    "Do a serial list repeatedly for timeout period"
    P.list_usb_w_timeout(r.get_in_config(["waiting", "timeout"]))


def get_usb_and_id():
    """
    Retrieve and set the USB device, the board name,  and the device id.
    Uses 'particle serial list' in a timeout loop. This is required
    for most things. Wait and handshake, use the usb device,
    and the id is needed by many things.
    """
    # PB['usb_device'], PB['device_id'] = P.get_usb_and_id()
    path, name, id = P.get_w_timeout(r.get_in_config(["waiting", "timeout"]))

    r.set({"device": {"path": path, "name": name, "id": id}})
    r.show()


def wait_for_plist():
    """Wait for particle serial list to succeed with timeout, doesn't
    really work."""
    # we don't care about the results, just to wait.
    P.get_w_timeout(r.get_in_config(["waiting", "timeout"]))


def archive_log():
    "Move the current logfile to one named after the current device."
    r.archive_log("%s.log" % r.get_in_device("id"))


def particle_help():
    """A function to provide additional Application specific help."""
    print(
        """\n\n Particle Process help:\n
Particle help can be accessed directly at any time with the 'particle-help'
command.\n
When working with a Particle.io board, the first step is to 'get'
the id and device. There are many commands here which need the usb device
or the device id.  With that, the basic commands for the process we have
been intending would be;

'start, setup, claim, testit, flash, and archive_log'.

Note that 'start', 'setup', and 'testit', are user defined commands which are
actually lists of other commands. These are defined in the configuration
file.  Help lists the source for these commands.
'start' is actually; 'dialog-start get wait identify'

The 'continue to next' and 'device failed' dialog windows are built in to
the interactive loop. Any additional prompts, such as 'dialog-start' or
'dialog-test' can be added to the process  with their commands.
and prompt texts are defined in the configuration, which can be
seen with 'showin config'.

When a process is deemed good, the autoexec can be set to it, and that
will be what runs automatically, in the process loop, or as a oneshot,
if no commands are given on the cli.\n\n """
    )


symbols = [
    ["dfu", P.dfu_mode, "Put the device in dfu mode."],
    ["listen", P.listen, "Put the device in listening mode."],
    ["list", P.list_usb, "List the particle boards connected to USB."],
    ["reset", reset_usb, "Reset the usb device"],
    ["identify", P.identify, "listen and identify"],
    ["inspect", P.inspect, "Inspect the device"],
    ["login", P.login, "Login to the particle cloud."],
    ["logout", P.logout, "Logout of the particle cloud."],
    ["update", P.update, "dfu/Update device, dfu then update"],
    ["setup-done", P.set_setup_bit, "Set the Setup bit."],
    ["add", add, "Add/claim device"],
    ["product-add", product_add, "Add device to a product"],
    ["claim", claim, "Cloud claim the device"],
    ["release", P.release, "Release claim on device"],
    ["flash", flash_image, "dfu/Flash the configured image"],
    ["tinker", flash_tinker, "dfu/Flash tinker onto device"],
    ["flash-test", flash_test, "Flash the test, only, the handshake is up to you."],
    ["_doctor", P.doctor, "run particle doctor"],
    ["cloud-status", cloud_status, "Get the cloud status of the current device"],
    ["get", get_usb_and_id, "Get the USB device and the device id."],
    ["archive-log", archive_log, "Archive the log to the the device id."],
    ["input-serial", input_serial, "Dialog to receive an 8 digit serial number."],
    ["particle-help", particle_help, "Additional Application layer help."],
]


# Name, function, number of args, help string
# Commands we want in the repl which can take arguments.
# look in rcore.py for examples.
specials = [
    [
        "_flash",
        P.flash,
        1,
        "Flash the specified image; _flash boron-system-part1@2.0.1.bin",
    ]
]


# get the default parser for the application and add to it if desired.
parser = None
# parser = r.get_parser()
# parser.add_argument("-f", "--foo", action="store_true", help="set foo")


def init():
    """
    Call into the interpreter/repl with our stuff,
    This starts everything up.
    """
    r.init(symbols, specials, parser)



import pathlib

# python 3.9 is required.
try:
    # Use setup() from setuptools(/distribute) if available
    from setuptools import setup
except ImportError:
    from distutils.core import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

from Particle_Board_REPL import __version__

setup(
    name="Particle_Board_REPL",
    version=__version__,
    # Your name & email here
    long_description=README,
    long_description_content_type="text/markdown",
    description="An application with REPL for programming and testing Particle.io boards.",
    url="https://github.com/EricGebhart/Particle_Board_REPL",
    author="Eric Gebhart",
    author_email="e.a.gebhart@gmail.com",
    packages=["Particle_Board_REPL"],
    package_data={"": ["*.yaml"]},
    include_package_data=True,
    scripts=[],
    license="MIT",
    entry_points={"console_scripts": ["PBR=Particle_Board_REPL.__main__:main"]},
    install_requires=["Simple_Process_REPL"],
)



#!/bin/python

import os
import subprocess
import logging
import time
import Simple_Process_REPL.core as r

logger = logging.getLogger()


def do_pcmd_w_timeout(cmd, timeout):
    """loop over a command for a timeout period. For commands
    that return failure quickly. Didn't work out for 'particle serial list',
    because it returns 0 no matter what happens."""
    start = time.time()
    while True:
        command = r.mk_cmd(cmd, prefix="particle")
        res = subprocess.run(command, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

        print(res)
        logger.debug(res)
        if res.returncode == 0:
            break

        if time.time() - start >= timeout:
            logger.warning("%s: timed out" % " ".join(command))
            break

        time.sleep(1)

    # print(res.stdout)
    logger.info(res.stdout.decode("utf-8"))
    return res.stdout.decode("utf-8")


def get_w_timeout(timeout):
    """
    loop over list_usb() for a timeout period. Extract
    the usb device, board name, and the device id if possible.
    This is actually nice behavior because it gives a chance to
    replug your particle board. And this is the first step to
    any set of operations on a device.
    """
    start = time.time()
    while True:
        stdout = list_usb()

        # USB, id
        if stdout.split(" ")[0] != "No":
            device_line = stdout.split("\n")[1]
            devices = device_line.split(" - ")
            return [devices[0], devices[1], devices[2]]

        if time.time() - start >= timeout:
            logger.error("Getting USB and device ID timed out")
            raise Exception("Getting USB and device ID timed out.")

        time.sleep(1)

    return stdout


def do_pcmd(cmd):
    """run a particle command, read and return it's output."""
    return r.do_cmd(r.mk_cmd(cmd, prefix="particle"))


# this doesn't actually work because particle returns 0
# no matter what happens.
def list_usb_w_timeout(timeout):
    "list the particle devices connected to usb."
    return do_pcmd_w_timeout("serial list", timeout)


def list_usb():
    "list the particle devices connected to usb."
    return do_pcmd("serial list")


def inspect():
    "Inspect the particle device"
    return do_pcmd("serial inspect")


def login():
    "Login to particle cloud from the cli prompt"
    return do_pcmd("cloud login")


def logout():
    "Logout of particle cloud, clean up tokens."
    return do_pcmd("cloud logout")


def get_usb_and_id():
    """
    Does a particle list and retrieves the usb device and device id
    Returns: usb, id
    """
    devices = list_usb()
    # USB, id
    return [devices.split(" - ")[2], devices.split()[0]]


def identify():
    """
    Get the identity of the particle board,
    send it to the log. (listen then identify)
    """
    logger.info("Identify: Start listening")
    listen()
    try:
        identify = do_pcmd("identify")
    except Exception as e:
        logger.error("Unable to Identify the particle board.")
        logger.error(e)
        raise (e)
    else:
        return identify


def doctor():
    """Run particle doctor. (list, dfu, doctor)"""
    list_usb()
    dfu_mode()
    return do_pcmd("-v doctor")


def _doctor():
    """Run particle doctor. (doctor)"""
    return do_pcmd("-v doctor")


def listen():
    """Start listening."""
    do_pcmd("usb start-listening")
    logger.info("Listening")


def update():
    """ Update the device OS. ie. dfu, update."""
    logger.info("Updating device")
    dfu_mode()
    return do_pcmd("update")


def set_setup_bit():
    """Mark the setup bit done."""
    do_pcmd("usb setup-done")
    logger.info("Set: Setup-done")


# no command product add...
# particle help product add --> gives nothing.
def add(device_id):
    """Register/claim device with 'particle device add'"""
    logger.info("Registering device:")
    # particle product add $product_id $device_id
    do_pcmd("device add {}".format(device_id))


def product_add(product, device_id):
    """Associate device with a product 'particle product device add'"""
    logger.info("Registering device:")
    do_pcmd("product device add %s %s" % (product, device_id))


def claim(device_id):
    """Cloud claim device with 'particle cloud claim'"""
    logger.info("Claiming device:")
    # particle product add $product_id $device_id
    do_pcmd("cloud claim {}".format(device_id))


def release(device_id):
    """Release the claim on a device"""
    do_pcmd("device remove {}".format(device_id))


def cloud_status(device_id):
    """Check the cloud-status of the device."""
    do_pcmd("usb cloud-status {}".format(device_id))


def reset_usb(device_id):
    """Reset the device on the usb."""
    do_pcmd("usb reset {}".format(device_id))


def flash(image):
    """flash an image. (dfu, flash --usb)"""
    logger.info("Flashing device with: %s" % image)
    if os.path.exists(image):
        dfu_mode()
        # do_pcmd('flash ––usb {}'.format(image))
        foo = os.popen("particle flash --usb %s" % image).read()
        logger.debug(foo)
    else:
        logger.error("Image %s to flash does not exist." % image)


def dfu_mode():
    """Put USB device in dfu mode."""
    do_pcmd("usb dfu")
